<?php
/*
LogSplitter: Logic class to split Squid/Pound access log files
(c)2006-2007 Tim Jackson (tim@timj.co.uk)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License version 2.1 as published by the Free Software Foundation.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
*/

class Text_LogSplitter
{
	var $_version = '0.2.1';

	var $log_fp;
	var $known_hostnames = array();
	var $config;
	var $known_hostnames_wildcard = array();
	
	var $known_hostnames_quicklist = array();
	
	var $output_fps = array();
	
	var $line_count = 0;
	var $bytes_processed = 0;
	
	var $_type = '';
	
	const LOGTYPE_SQUID = 'squid';
	const LOGTYPE_POUND = 'pound';
	
	function Text_LogSplitter ($config_filename)
	{
		$this->config = @parse_ini_file($config_filename);

		if ($this->config === array()) {
			$this->_error("Could not read config file '$config_filename'");
		}
		
		$this->log_fp = fopen($this->config['access_log'], 'r');
		
		if (!$this->log_fp) {
			$this->_error("Could not open access log file '".$this->config['access_log']."'");
		}
		
		$known_hostnames = $this->_getKnownHostnames();
		$this->_parseKnownHostnames($known_hostnames);
	}

	function _error($message)
	{
		// TODO: force PHP5 and throw exceptions
		print "ERROR: $message\n";
		die();
	}
	
	function run()
	{
		while ($line = trim(fgets($this->log_fp))) {
			$this->_processLine($line);
		}
		$this->_closeFiles();
	}
	
	// return array of known valid local hostnames
	// can include '*' wildcards e.g. *.example.com
	function _getKnownHostnames()
	{	
		$hosts = array();
		$lines = @file($this->config['hostname_list_file']);
		
		if ($lines === false) {
			$this->_throwError('Could not open hostname list file');
		}
		
		foreach ($lines as $line) {
			if (preg_match_all('/\s*(\S+)/', $line, $matches, PREG_PATTERN_ORDER)) {
				$base_hostname = $matches[0][0];
				for ($x=0; $x <= count($matches[0])-1; $x++) {
					$hosts[$base_hostname][] = trim($matches[0][$x]);
				}
			} else {
				$line_trimmed = trim($line);
				if (!empty($line_trimmed)) {
					print "Bad line in config: $line\n";
				}
			}
		}
		return $hosts;
	}
	
	function _parseKnownHostnames($hostnames)
	{
		foreach ($hostnames as $host_base => $host_list) {
			foreach ($host_list as $host) {
				if (strpos($host, '*') !== false) {
					$this->known_hostnames_wildcard[$host_base][] = str_replace('*','.*',str_replace('.','\\.', $host));
					//print "wildcard: $host";
				} else {
					$this->known_hostnames[$host_base] = $host;
					$this->known_hostnames_quicklist[$host] = $host_base;
				}
			}
		}
	}
	
	/**
	 * NB that this function canonicalises the hostname to lowercase
	 */
	function _processLine($line)
	{
		$this->line_count++;
		$this->bytes_processed += strlen($line);
		$orig_line = $line;
		//print "Process line: $line\n";
		# convert "GET http://localhost:8000/index.html => "GET /index.html
		$base_line = $this->_getBaseLine($line);
		if ($base_line !== false) {
			$hostname = strtolower($base_line['hostname']);
			$line = $base_line['line'];
			$matched = false;
			if (isset($this->known_hostnames_quicklist[$hostname])) {
				$matched = true;
				$matched_hostname = $hostname;
				$matched_base_hostname = $this->known_hostnames_quicklist[$hostname];
			} else {
				foreach ($this->known_hostnames_wildcard as $base_host => $host_list) {
					foreach ($host_list as $host) {
						//print "check $hostname against ^$host$\n";
						if (preg_match('/^'.$host.'$/', $hostname)) {
							//print "ok";
							$matched = true;
							$matched_hostname = str_replace('.*','*',stripslashes($host));
							$matched_base_hostname = $base_host;
						}
					}
				}
			}
			
			if ($matched) {
				//print "\tMatched $matched_hostname (base=$matched_base_hostname)\n";
			} else {
				//print "\tDid not match a known local hostname\n";
				$matched_base_hostname = 'unknown';
				$line = $orig_line;
			}
			$this->_outputLine($matched_base_hostname, $line);
			
		} else {
			print "Parse fail: $line\n";
			return false;
		}
	}
	
	/**
	 * Convert a raw log line to "base" format, that is "1.2.3.4 - - [30/Oct/2006:21:11:55 +0000] GET /foo/bar etc."
	 * This function REMOVES any port numbers from the log
	 */
	 //Oct 30 21:11:55 machine1 pound: www.example.com 1.2.3.4 - - [30/Oct/2006:21:11:55 +0000] "GET /test1 HTTP/1.0" 304 377 "-" "Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0) Active Cache Request"
	function _getBaseLine($line)
	{
		switch ($this->config['type']) {
			case self::LOGTYPE_POUND:
				if (preg_match('|pound: (\S+)(:\d+)? (.+)$|U', $line, $matches)) {
					$out['hostname'] = trim($matches[1]);
					// port is matches[2]
					$out['line'] = $matches[3];
					return $out;
				} else {
					return false;
				}
			case self::LOGTYPE_SQUID:
				if (preg_match('|"[A-Z]+ (\w+://)([\w\.-]*)(:[\d]+)?|', $line, $matches)) {
					// matches[1] is "http://" (if present)
					$out['hostname'] = trim($matches[2]);
					// matches[3] is ":80" (or similar) (if present)
					$out['line'] = str_replace($matches[1].$matches[2].$matches[3], '', $line);
					return $out;
				} else {
					return false;
				}
			default:
				$this->_throwError('Unknown log format type');
				
		}
		return false;
	}

	function _outputLine($base_host, $line)
	{
		if (!isset($this->output_fps[$base_host])) {
			$output_file = sprintf($this->config['split_output'], $base_host);
			$this->output_fps[$base_host] = fopen($output_file, 'w');
			if ($this->output_fps[$base_host] === false) {
				$this->_throwError("Could not open output log file $output_file");
			}
		}
		fputs($this->output_fps[$base_host], "$line\n");
	}

	function _closeFiles()
	{
		fclose($this->log_fp);
		foreach ($this->output_fps as $fp) {
			fclose($fp);
		}
	}
	
	function _throwError($msg)
	{
		print "$msg\n";
		die(5);
	}
	
	function getVersion()
	{
		return $this->_version;
	}
	
}
